@@ -33,7 +33,12 @@ from llm import (
 
 from .migrations import migrate
 from .plugins import pm, load_plugins
-from .utils import mimetype_from_path, mimetype_from_string, token_usage_string
+from .utils import (
+    mimetype_from_path,
+    mimetype_from_string,
+    token_usage_string,
+    extract_first_fenced_code_block,
+)
 import base64
 import httpx
 import pathlib
@@ -204,6 +209,7 @@ def cli():
 @click.option("--save", help="Save prompt with this template name")
 @click.option("async_", "--async", is_flag=True, help="Run prompt asynchronously")
 @click.option("-u", "--usage", is_flag=True, help="Show token usage")
+@click.option("-x", "--extract", is_flag=True, help="Extract first fenced code block")
 def prompt(
     prompt,
     system,
@@ -222,6 +228,7 @@ def prompt(
     save,
     async_,
     usage,
+    extract,
 ):
     """
     Execute a prompt
@@ -243,6 +250,12 @@ def prompt(
         cat image | llm 'describe image' -a -
         # With an explicit mimetype:
         cat image | llm 'describe image' --at - image/jpeg
+
+    The -x/--extract option returns just the content of the first ``` fenced code
+    block, if one is present. If none are present it returns the full response.
+
+    \b
+        llm 'JavaScript function for reversing a string' -x
     """
     if log and no_log:
         raise click.ClickException("--log and --no-log are mutually exclusive")
@@ -303,6 +316,8 @@ def prompt(
             to_save["system"] = system
         if param:
             to_save["defaults"] = dict(param)
+        if extract:
+            to_save["extract"] = True
         path.write_text(
             yaml.dump(
                 to_save,
@@ -319,6 +334,7 @@ def prompt(
         if system:
             raise click.ClickException("Cannot use -t/--template and --system together")
         template_obj = load_template(template)
+        extract = template_obj.extract
         prompt = read_prompt()
         try:
             prompt, system = template_obj.evaluate(prompt, params)
@@ -327,6 +343,9 @@ def prompt(
         if model_id is None and template_obj.model:
             model_id = template_obj.model
 
+    if extract:
+        no_stream = True
+
     conversation = None
     if conversation_id or _continue:
         # Load the conversation - loads most recent if no ID provided
@@ -407,7 +426,10 @@ def prompt(
                         system=system,
                         **validated_options,
                     )
-                    print(await response.text())
+                    text = await response.text()
+                    if extract:
+                        text = extract_first_fenced_code_block(text) or text
+                    print(text)
                 return response
 
             response = asyncio.run(inner())
@@ -424,8 +446,17 @@ def prompt(
                     sys.stdout.flush()
                 print("")
             else:
-                print(response.text())
+                text = response.text()
+                if extract:
+                    text = extract_first_fenced_code_block(text) or text
+                print(text)
+    # List of exceptions that should never be raised in pytest:
+    except (ValueError, NotImplementedError) as ex:
+        raise click.ClickException(str(ex))
     except Exception as ex:
+        # All other exceptions should raise in pytest, show to user otherwise
+        if getattr(sys, "_called_from_test", False):
+            raise
         raise click.ClickException(str(ex))
 
     if isinstance(response, AsyncResponse):
@@ -830,6 +861,7 @@ order by prompt_attachments."order"
 @click.option("-t", "--truncate", is_flag=True, help="Truncate long strings in output")
 @click.option("-u", "--usage", is_flag=True, help="Include token usage")
 @click.option("-r", "--response", is_flag=True, help="Just output the last response")
+@click.option("-x", "--extract", is_flag=True, help="Extract first fenced code block")
 @click.option(
     "current_conversation",
     "-c",
@@ -858,6 +890,7 @@ def logs_list(
     truncate,
     usage,
     response,
+    extract,
     current_conversation,
     conversation_id,
     json_output,
@@ -954,6 +987,7 @@ def logs_list(
                 else:
                     row[key] = json.loads(row[key])
 
+    output = None
     if json_output:
         # Output as JSON if requested
         for row in rows:
@@ -961,11 +995,20 @@ def logs_list(
                 {k: v for k, v in attachment.items() if k != "response_id"}
                 for attachment in attachments_by_id.get(row["id"], [])
             ]
-        click.echo(json.dumps(list(rows), indent=2))
+        output = json.dumps(list(rows), indent=2)
+    elif extract:
+        # Extract and return first code block
+        for row in rows:
+            output = extract_first_fenced_code_block(row["response"])
+            if output is not None:
+                break
     elif response:
         # Just output the last response
         if rows:
-            click.echo(rows[-1]["response"])
+            output = rows[-1]["response"]
+
+    if output is not None:
+        click.echo(output)
     else:
         # Output neatly formatted human-readable logs
         current_system = None
@@ -1051,12 +1094,15 @@ _type_lookup = {
     "--options", is_flag=True, help="Show options for each model, if available"
 )
 @click.option("async_", "--async", is_flag=True, help="List async models")
-def models_list(options, async_):
+@click.option("-q", "--query", help="Search for models matching this string")
+def models_list(options, async_, query):
     "List available models"
     models_that_have_shown_options = set()
     for model_with_aliases in get_models_with_aliases():
         if async_ and not model_with_aliases.async_model:
             continue
+        if query and not model_with_aliases.matches(query):
+            continue
         extra = ""
         if model_with_aliases.aliases:
             extra = " (aliases: {})".format(", ".join(model_with_aliases.aliases))
@@ -1096,6 +1142,8 @@ def models_list(options, async_):
             )
             output += "\n  Attachment types:\n{}".format(wrapper.fill(attachment_types))
         click.echo(output)
+    if not query:
+        click.echo(f"Default: {get_default_model()}")
 
 
 @models.command(name="default")
