@@ -1,6 +1,11 @@
-from llm import EmbeddingModel, Model, hookimpl
+from llm import AsyncModel, EmbeddingModel, Model, hookimpl
 import llm
-from llm.utils import dicts_to_table_string, remove_dict_none_values, logging_client
+from llm.utils import (
+    dicts_to_table_string,
+    remove_dict_none_values,
+    logging_client,
+    simplify_usage_dict,
+)
 import click
 import datetime
 import httpx
@@ -16,28 +21,68 @@ except ImportError:
     from pydantic.fields import Field
     from pydantic.class_validators import validator as field_validator  # type: ignore [no-redef]
 
-from typing import List, Iterable, Iterator, Optional, Union
+from typing import AsyncGenerator, List, Iterable, Iterator, Optional, Union
 import json
 import yaml
 
 
 @hookimpl
 def register_models(register):
-    register(Chat("gpt-3.5-turbo"), aliases=("3.5", "chatgpt"))
-    register(Chat("gpt-3.5-turbo-16k"), aliases=("chatgpt-16k", "3.5-16k"))
-    register(Chat("gpt-4"), aliases=("4", "gpt4"))
-    register(Chat("gpt-4-32k"), aliases=("4-32k",))
-    # GPT-4 Turbo models
-    register(Chat("gpt-4-1106-preview"))
-    register(Chat("gpt-4-0125-preview"))
-    register(Chat("gpt-4-turbo-2024-04-09"))
-    register(Chat("gpt-4-turbo"), aliases=("gpt-4-turbo-preview", "4-turbo", "4t"))
     # GPT-4o
-    register(Chat("gpt-4o", vision=True), aliases=("4o",))
-    register(Chat("gpt-4o-mini", vision=True), aliases=("4o-mini",))
+    register(
+        Chat("gpt-4o", vision=True), AsyncChat("gpt-4o", vision=True), aliases=("4o",)
+    )
+    register(
+        Chat("gpt-4o-mini", vision=True),
+        AsyncChat("gpt-4o-mini", vision=True),
+        aliases=("4o-mini",),
+    )
+    for audio_model_id in (
+        "gpt-4o-audio-preview",
+        "gpt-4o-audio-preview-2024-12-17",
+        "gpt-4o-audio-preview-2024-10-01",
+        "gpt-4o-mini-audio-preview",
+        "gpt-4o-mini-audio-preview-2024-12-17",
+    ):
+        register(
+            Chat(audio_model_id, audio=True),
+            AsyncChat(audio_model_id, audio=True),
+        )
+    # 3.5 and 4
+    register(
+        Chat("gpt-3.5-turbo"), AsyncChat("gpt-3.5-turbo"), aliases=("3.5", "chatgpt")
+    )
+    register(
+        Chat("gpt-3.5-turbo-16k"),
+        AsyncChat("gpt-3.5-turbo-16k"),
+        aliases=("chatgpt-16k", "3.5-16k"),
+    )
+    register(Chat("gpt-4"), AsyncChat("gpt-4"), aliases=("4", "gpt4"))
+    register(Chat("gpt-4-32k"), AsyncChat("gpt-4-32k"), aliases=("4-32k",))
+    # GPT-4 Turbo models
+    register(Chat("gpt-4-1106-preview"), AsyncChat("gpt-4-1106-preview"))
+    register(Chat("gpt-4-0125-preview"), AsyncChat("gpt-4-0125-preview"))
+    register(Chat("gpt-4-turbo-2024-04-09"), AsyncChat("gpt-4-turbo-2024-04-09"))
+    register(
+        Chat("gpt-4-turbo"),
+        AsyncChat("gpt-4-turbo"),
+        aliases=("gpt-4-turbo-preview", "4-turbo", "4t"),
+    )
     # o1
-    register(Chat("o1-preview", can_stream=False, allows_system_prompt=False))
-    register(Chat("o1-mini", can_stream=False, allows_system_prompt=False))
+    for model_id in ("o1", "o1-2024-12-17"):
+        register(
+            Chat(model_id, vision=True, can_stream=False),
+            AsyncChat(model_id, vision=True, can_stream=False),
+        )
+
+    register(
+        Chat("o1-preview", allows_system_prompt=False),
+        AsyncChat("o1-preview", allows_system_prompt=False),
+    )
+    register(
+        Chat("o1-mini", allows_system_prompt=False),
+        AsyncChat("o1-mini", allows_system_prompt=False),
+    )
     # The -instruct completion model
     register(
         Completion("gpt-3.5-turbo-instruct", default_max_tokens=256),
@@ -59,6 +104,9 @@ def register_models(register):
         api_version = extra_model.get("api_version")
         api_engine = extra_model.get("api_engine")
         headers = extra_model.get("headers")
+        kwargs = {}
+        if extra_model.get("can_stream") is False:
+            kwargs["can_stream"] = False
         if extra_model.get("completion"):
             klass = Completion
         else:
@@ -71,6 +119,7 @@ def register_models(register):
             api_version=api_version,
             api_engine=api_engine,
             headers=headers,
+            **kwargs,
         )
         if api_base:
             chat_model.needs_key = None
@@ -85,14 +134,39 @@ def register_models(register):
 @hookimpl
 def register_embedding_models(register):
     register(
-        OpenAIEmbeddingModel("ada-002", "text-embedding-ada-002"), aliases=("ada",)
+        OpenAIEmbeddingModel("text-embedding-ada-002", "text-embedding-ada-002"),
+        aliases=(
+            "ada",
+            "ada-002",
+        ),
+    )
+    register(
+        OpenAIEmbeddingModel("text-embedding-3-small", "text-embedding-3-small"),
+        aliases=("3-small",),
+    )
+    register(
+        OpenAIEmbeddingModel("text-embedding-3-large", "text-embedding-3-large"),
+        aliases=("3-large",),
     )
-    register(OpenAIEmbeddingModel("3-small", "text-embedding-3-small"))
-    register(OpenAIEmbeddingModel("3-large", "text-embedding-3-large"))
     # With varying dimensions
-    register(OpenAIEmbeddingModel("3-small-512", "text-embedding-3-small", 512))
-    register(OpenAIEmbeddingModel("3-large-256", "text-embedding-3-large", 256))
-    register(OpenAIEmbeddingModel("3-large-1024", "text-embedding-3-large", 1024))
+    register(
+        OpenAIEmbeddingModel(
+            "text-embedding-3-small-512", "text-embedding-3-small", 512
+        ),
+        aliases=("3-small-512",),
+    )
+    register(
+        OpenAIEmbeddingModel(
+            "text-embedding-3-large-256", "text-embedding-3-large", 256
+        ),
+        aliases=("3-large-256",),
+    )
+    register(
+        OpenAIEmbeddingModel(
+            "text-embedding-3-large-1024", "text-embedding-3-large", 1024
+        ),
+        aliases=("3-large-1024",),
+    )
 
 
 class OpenAIEmbeddingModel(EmbeddingModel):
@@ -146,8 +220,8 @@ def register_commands(cli):
             to_print = []
             for model in models:
                 # Print id, owned_by, root, created as ISO 8601
-                created_str = datetime.datetime.utcfromtimestamp(
-                    model["created"]
+                created_str = datetime.datetime.fromtimestamp(
+                    model["created"], datetime.timezone.utc
                 ).isoformat()
                 to_print.append(
                     {
@@ -248,18 +322,26 @@ class SharedOptions(llm.Options):
         return validated_logit_bias
 
 
-class Chat(Model):
-    needs_key = "openai"
-    key_env_var = "OPENAI_API_KEY"
+def _attachment(attachment):
+    url = attachment.url
+    base64_content = ""
+    if not url or attachment.resolve_type().startswith("audio/"):
+        base64_content = attachment.base64_content()
+        url = f"data:{attachment.resolve_type()};base64,{base64_content}"
+    if attachment.resolve_type().startswith("image/"):
+        return {"type": "image_url", "image_url": {"url": url}}
+    else:
+        format_ = "wav" if attachment.resolve_type() == "audio/wav" else "mp3"
+        return {
+            "type": "input_audio",
+            "input_audio": {
+                "data": base64_content,
+                "format": format_,
+            },
+        }
 
-    default_max_tokens = None
-
-    class Options(SharedOptions):
-        json_object: Optional[bool] = Field(
-            description="Output a valid JSON object {...}. Prompt must mention JSON.",
-            default=None,
-        )
 
+class _Shared:
     def __init__(
         self,
         model_id,
@@ -272,6 +354,7 @@ class Chat(Model):
         headers=None,
         can_stream=True,
         vision=False,
+        audio=False,
         allows_system_prompt=True,
     ):
         self.model_id = model_id
@@ -286,21 +369,31 @@ class Chat(Model):
         self.vision = vision
         self.allows_system_prompt = allows_system_prompt
 
+        self.attachment_types = set()
+
         if vision:
-            self.attachment_types = {
-                "image/png",
-                "image/jpeg",
-                "image/webp",
-                "image/gif",
-            }
+            self.attachment_types.update(
+                {
+                    "image/png",
+                    "image/jpeg",
+                    "image/webp",
+                    "image/gif",
+                }
+            )
+
+        if audio:
+            self.attachment_types.update(
+                {
+                    "audio/wav",
+                    "audio/mpeg",
+                }
+            )
 
     def __str__(self):
         return "OpenAI Chat: {}".format(self.model_id)
 
-    def execute(self, prompt, stream, response, conversation=None):
+    def build_messages(self, prompt, conversation):
         messages = []
-        if prompt.system and not self.allows_system_prompt:
-            raise NotImplementedError("Model does not support system prompts")
         current_system = None
         if conversation is not None:
             for prev_response in conversation.responses:
@@ -313,70 +406,45 @@ class Chat(Model):
                     )
                     current_system = prev_response.prompt.system
                 if prev_response.attachments:
-                    attachment_message = [
-                        {"type": "text", "text": prev_response.prompt.prompt}
-                    ]
-                    for attachment in prev_response.attachments:
-                        url = attachment.url
-                        if not url:
-                            base64_image = attachment.base64_content()
-                            url = f"data:{attachment.resolve_type()};base64,{base64_image}"
+                    attachment_message = []
+                    if prev_response.prompt.prompt:
                         attachment_message.append(
-                            {"type": "image_url", "image_url": {"url": url}}
+                            {"type": "text", "text": prev_response.prompt.prompt}
                         )
+                    for attachment in prev_response.attachments:
+                        attachment_message.append(_attachment(attachment))
                     messages.append({"role": "user", "content": attachment_message})
                 else:
                     messages.append(
                         {"role": "user", "content": prev_response.prompt.prompt}
                     )
-                messages.append({"role": "assistant", "content": prev_response.text()})
+                messages.append(
+                    {"role": "assistant", "content": prev_response.text_or_raise()}
+                )
         if prompt.system and prompt.system != current_system:
             messages.append({"role": "system", "content": prompt.system})
         if not prompt.attachments:
             messages.append({"role": "user", "content": prompt.prompt})
         else:
-            attachment_message = [{"type": "text", "text": prompt.prompt}]
+            attachment_message = []
+            if prompt.prompt:
+                attachment_message.append({"type": "text", "text": prompt.prompt})
             for attachment in prompt.attachments:
-                url = attachment.url
-                if not url:
-                    base64_image = attachment.base64_content()
-                    url = f"data:{attachment.resolve_type()};base64,{base64_image}"
-                attachment_message.append(
-                    {"type": "image_url", "image_url": {"url": url}}
-                )
+                attachment_message.append(_attachment(attachment))
             messages.append({"role": "user", "content": attachment_message})
+        return messages
 
-        kwargs = self.build_kwargs(prompt, stream)
-        client = self.get_client()
-        if stream:
-            completion = client.chat.completions.create(
-                model=self.model_name or self.model_id,
-                messages=messages,
-                stream=True,
-                **kwargs,
-            )
-            chunks = []
-            for chunk in completion:
-                chunks.append(chunk)
-                try:
-                    content = chunk.choices[0].delta.content
-                except IndexError:
-                    content = None
-                if content is not None:
-                    yield content
-            response.response_json = remove_dict_none_values(combine_chunks(chunks))
-        else:
-            completion = client.chat.completions.create(
-                model=self.model_name or self.model_id,
-                messages=messages,
-                stream=False,
-                **kwargs,
-            )
-            response.response_json = remove_dict_none_values(completion.model_dump())
-            yield completion.choices[0].message.content
-        response._prompt_json = redact_data_urls({"messages": messages})
+    def set_usage(self, response, usage):
+        if not usage:
+            return
+        input_tokens = usage.pop("prompt_tokens")
+        output_tokens = usage.pop("completion_tokens")
+        usage.pop("total_tokens")
+        response.set_usage(
+            input=input_tokens, output=output_tokens, details=simplify_usage_dict(usage)
+        )
 
-    def get_client(self):
+    def get_client(self, async_=False):
         kwargs = {}
         if self.api_base:
             kwargs["base_url"] = self.api_base
@@ -396,7 +464,10 @@ class Chat(Model):
             kwargs["default_headers"] = self.headers
         if os.environ.get("LLM_OPENAI_SHOW_RESPONSES"):
             kwargs["http_client"] = logging_client()
-        return openai.OpenAI(**kwargs)
+        if async_:
+            return openai.AsyncOpenAI(**kwargs)
+        else:
+            return openai.OpenAI(**kwargs)
 
     def build_kwargs(self, prompt, stream):
         kwargs = dict(not_nulls(prompt.options))
@@ -410,6 +481,110 @@ class Chat(Model):
         return kwargs
 
 
+class Chat(_Shared, Model):
+    needs_key = "openai"
+    key_env_var = "OPENAI_API_KEY"
+    default_max_tokens = None
+
+    class Options(SharedOptions):
+        json_object: Optional[bool] = Field(
+            description="Output a valid JSON object {...}. Prompt must mention JSON.",
+            default=None,
+        )
+
+    def execute(self, prompt, stream, response, conversation=None):
+        if prompt.system and not self.allows_system_prompt:
+            raise NotImplementedError("Model does not support system prompts")
+        messages = self.build_messages(prompt, conversation)
+        kwargs = self.build_kwargs(prompt, stream)
+        client = self.get_client()
+        usage = None
+        if stream:
+            completion = client.chat.completions.create(
+                model=self.model_name or self.model_id,
+                messages=messages,
+                stream=True,
+                **kwargs,
+            )
+            chunks = []
+            for chunk in completion:
+                chunks.append(chunk)
+                if chunk.usage:
+                    usage = chunk.usage.model_dump()
+                try:
+                    content = chunk.choices[0].delta.content
+                except IndexError:
+                    content = None
+                if content is not None:
+                    yield content
+            response.response_json = remove_dict_none_values(combine_chunks(chunks))
+        else:
+            completion = client.chat.completions.create(
+                model=self.model_name or self.model_id,
+                messages=messages,
+                stream=False,
+                **kwargs,
+            )
+            usage = completion.usage.model_dump()
+            response.response_json = remove_dict_none_values(completion.model_dump())
+            yield completion.choices[0].message.content
+        self.set_usage(response, usage)
+        response._prompt_json = redact_data({"messages": messages})
+
+
+class AsyncChat(_Shared, AsyncModel):
+    needs_key = "openai"
+    key_env_var = "OPENAI_API_KEY"
+    default_max_tokens = None
+
+    class Options(SharedOptions):
+        json_object: Optional[bool] = Field(
+            description="Output a valid JSON object {...}. Prompt must mention JSON.",
+            default=None,
+        )
+
+    async def execute(
+        self, prompt, stream, response, conversation=None
+    ) -> AsyncGenerator[str, None]:
+        if prompt.system and not self.allows_system_prompt:
+            raise NotImplementedError("Model does not support system prompts")
+        messages = self.build_messages(prompt, conversation)
+        kwargs = self.build_kwargs(prompt, stream)
+        client = self.get_client(async_=True)
+        usage = None
+        if stream:
+            completion = await client.chat.completions.create(
+                model=self.model_name or self.model_id,
+                messages=messages,
+                stream=True,
+                **kwargs,
+            )
+            chunks = []
+            async for chunk in completion:
+                if chunk.usage:
+                    usage = chunk.usage.model_dump()
+                chunks.append(chunk)
+                try:
+                    content = chunk.choices[0].delta.content
+                except IndexError:
+                    content = None
+                if content is not None:
+                    yield content
+            response.response_json = remove_dict_none_values(combine_chunks(chunks))
+        else:
+            completion = await client.chat.completions.create(
+                model=self.model_name or self.model_id,
+                messages=messages,
+                stream=False,
+                **kwargs,
+            )
+            response.response_json = remove_dict_none_values(completion.model_dump())
+            usage = completion.usage.model_dump()
+            yield completion.choices[0].message.content
+        self.set_usage(response, usage)
+        response._prompt_json = redact_data({"messages": messages})
+
+
 class Completion(Chat):
     class Options(SharedOptions):
         logprobs: Optional[int] = Field(
@@ -466,7 +641,7 @@ class Completion(Chat):
             )
             response.response_json = remove_dict_none_values(completion.model_dump())
             yield completion.choices[0].text
-        response._prompt_json = redact_data_urls({"messages": messages})
+        response._prompt_json = redact_data({"messages": messages})
 
 
 def not_nulls(data) -> dict:
@@ -484,7 +659,7 @@ def combine_chunks(chunks: List) -> dict:
 
     for item in chunks:
         if item.usage:
-            usage = dict(item.usage)
+            usage = item.usage.dict()
         for choice in item.choices:
             if choice.logprobs and hasattr(choice.logprobs, "top_logprobs"):
                 logprobs.append(
@@ -512,18 +687,21 @@ def combine_chunks(chunks: List) -> dict:
     }
     if logprobs:
         combined["logprobs"] = logprobs
-    for key in ("id", "object", "model", "created", "index"):
-        value = getattr(chunks[0], key, None)
-        if value is not None:
-            combined[key] = value
+    if chunks:
+        for key in ("id", "object", "model", "created", "index"):
+            value = getattr(chunks[0], key, None)
+            if value is not None:
+                combined[key] = value
 
     return combined
 
 
-def redact_data_urls(input_dict):
+def redact_data(input_dict):
     """
     Recursively search through the input dictionary for any 'image_url' keys
     and modify the 'url' value to be just 'data:...'.
+
+    Also redact input_audio.data keys
     """
     if isinstance(input_dict, dict):
         for key, value in input_dict.items():
@@ -534,9 +712,11 @@ def redact_data_urls(input_dict):
                 and value["url"].startswith("data:")
             ):
                 value["url"] = "data:..."
+            elif key == "input_audio" and isinstance(value, dict) and "data" in value:
+                value["data"] = "..."
             else:
-                redact_data_urls(value)
+                redact_data(value)
     elif isinstance(input_dict, list):
         for item in input_dict:
-            redact_data_urls(item)
+            redact_data(item)
     return input_dict
